# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__doc__ = """

Helper functions for use with asynq (similar to itertools).

"""

from .contexts import AsyncContext
from .decorators import asynq, async_proxy, async_call, AsyncDecorator, AsyncDecoratorBinder
from .futures import ConstFuture
# we shouldn't use the return syntax in generators here so that asynq can be imported
# under Python versions that lack our patch to allow returning from generators
from .utils import result

from qcore import get_original_fn, utime
from qcore.caching import get_args_tuple, get_kwargs_defaults, LRUCache
from qcore.events import EventHook
from qcore.errors import reraise, prepare_for_reraise
from qcore.decorators import decorate
import functools
import inspect2
import itertools
import weakref
import threading


@asynq()
def amap(function, sequence):
    """Equivalent of map() that takes an async map function.

    Returns a list.

    """
    result((yield [function.asynq(elt) for elt in sequence])); return


@asynq()
def afilter(function, sequence):
    """Equivalent of filter() that takes an async filter function.

    Returns a list.

    """
    if function is None:
        result(filter(None, sequence)); return
    should_include = yield [function.asynq(elt) for elt in sequence]
    result(list(itertools.compress(sequence, should_include))); return


@asynq()
def afilterfalse(function, sequence):
    """Equivalent of itertools.ifilterfalse() that takes an async filter function.

    Returns a list.

    """
    should_exclude = yield [function.asynq(elt) for elt in sequence]
    should_include = [not res for res in should_exclude]
    result(list(itertools.compress(sequence, should_include))); return


@asynq()
def asorted(iterable, key=None, reverse=False):
    """Equivalent of sorted() that takes an async key function.

    The cmp= argument of sorted() is not supported.

    Returns a list.

    """
    values = list(iterable)
    if key is None:
        keys = values
    else:
        keys = yield amap.asynq(key, values)
    # we need to use key= here because otherwise we will compare the values when the key are
    # equal, which would be a behavior difference between sorted() and asorted()
    pairs = sorted(zip(keys, values), key=lambda p: p[0], reverse=reverse)
    result([p[1] for p in pairs]); return


@asynq()
def amax(*args, **kwargs):
    """Async equivalent of max()."""
    key_fn = kwargs.pop('key', None)
    if kwargs:
        raise TypeError('amax() got an unexpected keyword argument')

    if len(args) == 0:
        raise TypeError('amax() expected 1 arguments, got 0')
    elif len(args) == 1:
        iterable = args[0]
    else:
        iterable = args

    if key_fn is None:
        result(max(iterable)); return

    # support generators
    if not isinstance(iterable, (list, tuple)):
        iterable = list(iterable)

    keys = yield amap.asynq(key_fn, iterable)
    max_pair = max(enumerate(iterable), key=lambda pair: keys[pair[0]])
    result(max_pair[1]); return


@asynq()
def amin(*args, **kwargs):
    """Async equivalent of min()."""
    key_fn = kwargs.pop('key', None)
    if kwargs:
        raise TypeError('amin() got an unexpected keyword argument')

    if len(args) == 0:
        raise TypeError('amin() expected 1 arguments, got 0')
    elif len(args) == 1:
        iterable = args[0]
    else:
        iterable = args

    if key_fn is None:
        result(min(iterable)); return

    # support generators
    if not isinstance(iterable, (list, tuple)):
        iterable = list(iterable)

    keys = yield amap.asynq(key_fn, iterable)
    max_pair = min(enumerate(iterable), key=lambda pair: keys[pair[0]])
    result(max_pair[1]); return


@asynq()
def asift(pred, items):
    """Sifts a list of items into those that meet the predicate and those that don't."""
    yes = []
    no = []
    results = yield [pred.asynq(item) for item in items]
    for item, yesno in zip(items, results):
        if yesno:
            yes.append(item)
        else:
            no.append(item)
    result((yes, no)); return


def acached_per_instance():
    """Async equivalent of qcore.caching.cached_per_instance().

    Unlike l0cache, the cached value is stored in the instance so that it gets
    garbage collected together with the instance.

    The cached values are not stored when the object is pickled.

    """
    def cache_fun(fun):
        argspec = inspect2.getfullargspec(get_original_fn(fun))
        arg_names = argspec.args[1:] + argspec.kwonlyargs  # remove self
        async_fun = fun.asynq
        kwargs_defaults = get_kwargs_defaults(argspec)
        cache = {}

        def cache_key(args, kwargs):
            return get_args_tuple(args, kwargs, arg_names, kwargs_defaults)

        def clear_cache(instance_key, ref):
            del cache[instance_key]

        @asynq()
        @functools.wraps(fun)
        def new_fun(self, *args, **kwargs):
            instance_key = id(self)
            if instance_key not in cache:
                ref = weakref.ref(self, functools.partial(clear_cache, instance_key))
                cache[instance_key] = (ref, {})
            instance_cache = cache[instance_key][1]

            k = cache_key(args, kwargs)
            try:
                result(instance_cache[k]); return
            except KeyError:
                value = yield async_fun(self, *args, **kwargs)
                instance_cache[k] = value
                result(value); return

        # just so unit tests can check that this is cleaned up correctly
        new_fun.__acached_per_instance_cache__ = cache
        return new_fun
    return cache_fun


def alru_cache(maxsize=128, key_fn=None):
    """Async equivalent of qcore.caching.lru_cache().

    maxsize is the number of different keys cache can accomodate.
    key_fn is the function that builds key from args. The default key function
    creates a tuple out of args and kwargs. If you use the default it acts the same
    as functools.lru_cache (except with async).

    Possible use cases of key_fn:
    - Your cache key is very large, so you don't want to keep the whole key in memory.
    - The function takes some arguments that don't affect the result.

    """

    def decorator(fn):
        cache = LRUCache(maxsize)
        argspec = inspect2.getfullargspec(get_original_fn(fn))
        arg_names = argspec.args[1:] + argspec.kwonlyargs  # remove self
        async_fun = fn.asynq
        kwargs_defaults = get_kwargs_defaults(argspec)

        cache_key = key_fn
        if cache_key is None:
            def cache_key(args, kwargs):
                return get_args_tuple(args, kwargs, arg_names, kwargs_defaults)

        @asynq()
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            key = cache_key(args, kwargs)
            try:
                result(cache[key]); return
            except KeyError:
                value = yield async_fun(*args, **kwargs)
                cache[key] = value
                result(value); return

        return wrapper
    return decorator


@asynq()
def call_with_context(context, fn, *args, **kwargs):
    """Calls fn in the given with context.

    This is useful if you need to call two functions at once, but only one should be called in the
    context. For example:

        important, not_important = yield (
            get_important.asynq(oid),
            call_with_context.asynq(a.livenode.dep.IgnoreDependencies(), get_not_important, oid).
        )

    """
    with context:
        result((yield fn.asynq(*args, **kwargs))); return


class DeduplicateDecoratorBinder(AsyncDecoratorBinder):
    def dirty(self, *args, **kwargs):
        if self.instance is None:
            self.decorator.dirty(*args, **kwargs)
        else:
            self.decorator.dirty(self.instance, *args, **kwargs)


class DeduplicateDecorator(AsyncDecorator):
    binder_cls = DeduplicateDecoratorBinder
    tasks = {}

    def __init__(self, fn, task_cls, keygetter):
        AsyncDecorator.__init__(self, fn, task_cls)
        self.keygetter = keygetter

    def cache_key(self, args, kwargs):
        return self.keygetter(args, kwargs), threading.current_thread(), id(self.fn)

    def asynq(self, *args, **kwargs):
        cache_key = self.cache_key(args, kwargs)

        try:
            return self.tasks[cache_key]
        except KeyError:
            task = self.fn.asynq(*args, **kwargs)

            def callback(task):
                self.tasks.pop(cache_key, None)

            self.tasks[cache_key] = task
            task.on_computed.subscribe(callback)
            return task

    locals()['async'] = asynq

    def dirty(self, *args, **kwargs):
        cache_key = self.cache_key(args, kwargs)
        self.tasks.pop(cache_key, None)


def deduplicate(keygetter=None):
    """Decorator that ensures that no two identical instances of a task run concurrently.

    This is useful in situations like this:

        @asynq()
        def should_filter_object(oid, uid):
            data = yield get_data_for_user.asynq(uid)
            ...

        @asynq()
        def filter_objects(oids, uid):
            ... = yield [should_filter_object.asynq(oid, uid) for oid in oids]

    where get_data_for_user is cached (e.g. in memcache or l0cache). Without the deduplicate
    decorator, this may end up calling the body of the get_data_for_user function multiple times,
    despite the caching, because a second async task may enter the body while the first one is
    still active.

    You can also call dirty on a deduplicated function to remove a cached async task with the
    corresponding args and kwargs. This is useful if a deduplicating function ends up calling
    itself with the same args and kwargs, either directly or deeper in the call stack.

    """
    def decorator(fun):
        _keygetter = keygetter
        if _keygetter is None:
            original_fn = get_original_fn(fun)
            argspec = inspect2.getfullargspec(original_fn)
            arg_names = argspec.args + argspec.kwonlyargs
            kwargs_defaults = get_kwargs_defaults(argspec)
            _keygetter = lambda args, kwargs: get_args_tuple(args, kwargs, arg_names, kwargs_defaults)

        return decorate(DeduplicateDecorator, fun.task_cls, _keygetter)(fun)
    return decorator


class AsyncTimer(AsyncContext):
    """Simple async-aware timer class.

    Use this to find out how long a block of code takes within an async task. If
    other tasks run interspersed with the task in which this is used, time spent
    executing those tasks will not be counted. The result (in microseconds) will
    be available as the total_time attribute on the context object after exiting
    the context.

    The total_time attribute may have a nonzero value during the context if any
    yields were performed. Because of optimizations that may be done in the
    future to how contexts work between tasks, the value shouldn't be trusted
    until exiting the context.

    Usage example:
        @asynq()
        def potentially_slow_function(x):

            with AsyncTimer() as t:
                yield do_a_lot_of_work.asynq(x)
                # don't use t.total_time here!

            report_time_for_x(x, t.total_time)

        yield [potentially_slow_function(x) for x in all_x_values]

    """

    def __init__(self):
        self.total_time = 0
        self._last_start_time = None

    def resume(self):
        self._last_start_time = utime()

    def pause(self):
        self.total_time += utime() - self._last_start_time


class AsyncEventHook(EventHook):
    """EventHook that supports async handlers.

    When the event triggers, all the async handlers will be invoked asynchronously.

    All non-async handlers will be invoked normally (same as EventHook).

    """

    @asynq()
    def trigger(self, *args):
        yield [async_call.asynq(handler, *args) for handler in self]

    @asynq()
    def safe_trigger(self, *args):
        wrapped_handlers = [self._create_safe_wrapper(handler) for handler in self]
        results = yield [wrapped_handler.asynq(*args) for wrapped_handler in wrapped_handlers]
        for error in filter(None, results):
            reraise(error)

    @staticmethod
    def _create_safe_wrapper(handler):
        @asynq()
        def wrapped(*args):
            error = None
            try:
                yield async_call.asynq(handler, *args)
            except BaseException as e:
                prepare_for_reraise(e)
                error = e
            result(error); return
        return wrapped



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__doc__ = """

Async generators.

"""
import qcore
import functools

from .decorators import asynq, async_proxy
from .futures import ConstFuture
from .utils import result

END_OF_GENERATOR = qcore.MarkerObject(u'end of generator')


def async_generator():
    """Decorator to create an async generator.

    async functions are always implemented as generators, but it is sometimes useful to yield values
    (as in a normal, non-async generator) from an async function. This decorator provides that
    capability. Inside of the generator, wrap any data (as opposed to async futures) you want to
    yield in a Value object. For example:

    >>> @asynq()
    ... def async_function():
    ...     return 42
    >>> @async_generator()
    ... def gen():
    ...     value = yield async_function.asynq()
    ...     yield Value(value)
    >>> list_of_generator(gen())
    >>> [42]

    Async generators cannot function exactly like normal generators, because to do so their .next
    method would have to be invoked asynchronously, which is not possible in Python 2. (Python 3
    adds an async iteration protocol.) Therefore, async generators produce asynq Futures instead of
    producing their values directly. You should normally iterate over an async generator with code
    like:

        for task in generator():
            value = yield task
            # do stuff with value

    Additionally, generators that have an async yield after the last Value they yield will produce
    an additional value, END_OF_GENERATOR, that should be ignored by calling code. To guard against
    this possibility, you need to write:

        for task in generator():
            value = yield task
            if value is END_OF_GENERATOR:
                continue
            # do stuff with value

    """
    def decorator(fun):
        @functools.wraps(fun)
        def wrapped(*args, **kwargs):
            gen = fun(*args, **kwargs)
            return _AsyncGenerator(gen)
        return wrapped
    return decorator


class Value(object):
    """Represents a value yielded by an async generator."""
    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return '<Value: %r>' % self.value


@asynq()
def list_of_generator(generator):
    """Returns a list of all the values in the async generator."""
    data = []
    for task in generator:
        value = yield task
        if value is END_OF_GENERATOR:
            continue
        data.append(value)
    result(data); return


@asynq()
def take_first(generator, n):
    """Returns the first n values in the generator."""
    ret = []
    for i, task in enumerate(generator):
        value = yield task
        if value is END_OF_GENERATOR:
            continue
        ret.append(value)
        if i == n - 1:
            break
    result(ret); return


class _AsyncGenerator(object):
    def __init__(self, generator):
        self.generator = generator
        self.last_task = None
        self.is_stopped = False

    def __iter__(self):
        return self

    @async_proxy(pure=True)
    def next(self):
        return self.send.asynq(None)

    __next__ = next  # Python 3

    @async_proxy()
    def send(self, value):
        # ensure that the previous yielded value has been computed
        if self.last_task is not None:
            if not self.last_task.is_computed():
                raise RuntimeError(
                    'You must compute the previous task before advancing the generator')
        if self.is_stopped:
            raise StopIteration

        # get a first value so that we can raise StopIteration here immediately
        # this works only if there are no async yields after the last Value is yielded; otherwise
        # the generator will yield END_OF_GENERATOR before being exhausted
        first_value = self._get_one_value(value)
        if isinstance(first_value, Value):
            return ConstFuture(first_value.value)
        task = self._send_inner.asynq(first_value)
        self.last_task = task
        return task

    @asynq()
    def _send_inner(self, first_task):
        # ensure previous item has been consumed
        yield_result = yield first_task
        while True:
            try:
                value = self._get_one_value(yield_result)
            except StopIteration:
                result(END_OF_GENERATOR); return
            if isinstance(value, Value):
                result(value.value); return
            else:
                yield_result = yield value

    def _get_one_value(self, value):
        try:
            return self.generator.send(value)
        except StopIteration:
            self.is_stopped = True
            raise

    def __repr__(self):
        return '<@async_generator() %s %s>' % (self.generator, 'stopped' if self.stopped else '')



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import time

from asynq import asynq, AsyncContext, result, scheduler
from asynq.tools import (
    amap,
    afilter,
    afilterfalse,
    amin,
    amax,
    asift,
    asorted,
    acached_per_instance,
    alru_cache,
    call_with_context,
    deduplicate,
    AsyncTimer,
    AsyncEventHook,
)
from qcore.asserts import assert_eq, assert_gt, assert_is, AssertRaises, assert_unordered_list_eq


@asynq()
def inner_fn(x):
    pass


@asynq()
def filter_fn(elt):
    yield inner_fn.asynq(elt)
    result(elt is not None); return


@asynq()
def alen(seq):
    return len(seq)


def test_afilter():
    assert_eq([], list(afilter.asynq(filter_fn, []).value()))
    assert_eq([], list(afilter.asynq(filter_fn, [None]).value()))
    assert_eq([1], list(afilter(filter_fn, [None, 1, None])))
    assert_eq([1], list(afilter.asynq(filter_fn, [None, 1, None]).value()))
    assert_eq([1], list(afilter(None, [None, 1, None])))


def test_afilterfalse():
    assert_eq([], list(afilterfalse.asynq(filter_fn, []).value()))
    assert_eq([None], list(afilterfalse.asynq(filter_fn, [None]).value()))
    assert_eq([None, None], list(afilterfalse(filter_fn, [None, 1, None])))
    assert_eq([None, None], list(afilterfalse.asynq(filter_fn, [None, 1, None]).value()))


def test_asift():
    assert_eq(([], []), asift.asynq(filter_fn, []).value())
    assert_eq(([], [None]), asift.asynq(filter_fn, [None]).value())
    assert_eq(([1], [None, None]), asift(filter_fn, [None, 1, None]))
    assert_eq(([1], [None, None]), asift.asynq(filter_fn, [None, 1, None]).value())


def test_amap():
    assert_eq([False], list(amap(filter_fn, [None])))
    assert_eq([True], list(amap(filter_fn, [4])))
    assert_eq([], list(amap(filter_fn, [])))
    assert_eq([False, True, False], list(amap(filter_fn, [None, '', None])))


def test_asorted():
    assert_eq([], asorted([], key=filter_fn))
    assert_eq([None], asorted([None], key=filter_fn))
    assert_eq([None, True], asorted([True, None], key=filter_fn))
    assert_eq([1, 2], asorted([2, 1]))


def test_amax():
    assert_eq(1, amax(1, None, key=filter_fn))
    assert_eq(1, amax([1, None], key=filter_fn))
    assert_eq(1, amax((elt for elt in (1, None)), key=filter_fn))
    assert_eq([1, 2, 3], amax([1], [1, 2, 3], [1, 2], key=alen))
    assert_eq([1, 2, 3], amax([[1], [1, 2, 3], [1, 2]], key=alen))

    assert_eq(4, amax(1, 2, 3, 4))

    with AssertRaises(TypeError):
        amax(key=filter_fn)
    with AssertRaises(ValueError):
        amax([], key=filter_fn)
    with AssertRaises(TypeError):
        amax([], key=filter_fn, random_keyword_argument='raising a TypeError')


def test_amin():
    assert_is(None, amin(1, None, key=filter_fn))
    assert_is(None, amin([1, None], key=filter_fn))
    assert_is(None, amin((elt for elt in (1, None)), key=filter_fn))
    assert_eq([1], amin([1], [1, 2, 3], [1, 2], key=alen))
    assert_eq([1], amin([[1], [1, 2, 3], [1, 2]], key=alen))

    assert_eq(1, amin(1, 2, 3, 4))

    with AssertRaises(TypeError):
        amin(key=filter_fn)
    with AssertRaises(ValueError):
        amin([], key=filter_fn)
    with AssertRaises(TypeError):
        amin([], key=filter_fn, random_keyword_argument='raising a TypeError')


class AsyncObject(object):
    cls_value = 0

    def __init__(self):
        self.value = 0

    @acached_per_instance()
    @asynq()
    def get_value(self, index):
        self.value += 1
        return self.value

    @acached_per_instance()
    @asynq()
    def with_kwargs(self, x=1, y=2, z=3):
        self.value += (x + y + z)
        return self.value

    @acached_per_instance()
    @asynq()
    def raises_exception(self):
        assert False

    if sys.version_info >= (3, 0):
        exec("""
@acached_per_instance()
@asynq()
def with_kwonly_arg(self, *, arg=1):
    return arg
""")

    @deduplicate()
    @asynq()
    def increment_value_method(self, val=1):
        self.value += val

    @deduplicate()
    @asynq()
    @staticmethod
    def deduplicated_static_method(val=1):
        AsyncObject.cls_value += val


class UnhashableAcached(AsyncObject):
    __hash__ = None


def test_acached_per_instance():
    for cls in (AsyncObject, UnhashableAcached):
        obj = cls()
        cache = type(obj).get_value.decorator.__acached_per_instance_cache__
        assert_eq(0, len(cache), extra=repr(cache))

        assert_eq(1, obj.get_value(0))
        assert_eq(1, obj.get_value(0))
        assert_eq(2, obj.get_value(1))
        assert_eq(1, obj.get_value(0))
        assert_eq(1, obj.get_value(index=0))
        assert_eq(1, obj.get_value.asynq(index=0).value())

        assert_eq(8, obj.with_kwargs())
        assert_eq(8, obj.with_kwargs(z=3))
        assert_eq(17, obj.with_kwargs(x=3, y=3))

        assert_eq(1, len(cache), extra=repr(cache))

        if sys.version_info >= (3, 0):
            assert_eq(1, obj.with_kwonly_arg(arg=1))

        del obj
        assert_eq(0, len(cache), extra=repr(cache))


def test_acached_per_instance_exception_handling():
    obj = AsyncObject()
    try:
        obj.raises_exception()
    except AssertionError:
        # the exception should not affect the internals of the scheduler, and the active task
        # should get cleaned up
        assert_is(None, scheduler.get_active_task())


def test_alru_cache():
    _check_alru_cache()


@asynq()
def _check_alru_cache():
    @alru_cache(maxsize=1, key_fn=lambda args, kwargs: args[0] % 2 == 0)
    @asynq()
    def cube(n):
        return n * n * n

    assert_eq(1, (yield cube.asynq(1)))
    # hit the cache
    assert_eq(1, (yield cube.asynq(3)))
    # cache miss
    assert_eq(8, (yield cube.asynq(2)))
    # now it's a cache miss
    assert_eq(27, (yield cube.asynq(3)))


class Ctx(AsyncContext):
    is_on = False

    def pause(self):
        Ctx.is_on = False

    def resume(self):
        Ctx.is_on = True


@asynq()
def assert_state(value):
    yield AsyncObject().get_value.asynq(value)
    assert_is(value, Ctx.is_on)


def test_call_with_context():
    assert_state(False)
    call_with_context(Ctx(), assert_state, True)

i = 0


@deduplicate()
@asynq()
def increment_value(val=1):
    global i
    i += val


@deduplicate()
@asynq()
def recursive_incrementer(n):
    if n == 0:
        result((yield increment_value.asynq(n))); return
    result(recursive_incrementer(n - 1)); return


@deduplicate()
@asynq()
def call_with_dirty():
    call_with_dirty.dirty()


@deduplicate()
@asynq()
def recursive_call_with_dirty():
    global i
    if i > 0:
        result(i); return
    i += 1
    recursive_call_with_dirty.dirty()
    yield recursive_call_with_dirty.asynq()


if sys.version_info >= (3, 0):
    exec("""
@deduplicate()
@asynq()
def call_with_kwonly_arg(*, arg):
    return arg
""")


def test_deduplicate():
    _check_deduplicate()


@asynq()
def _check_deduplicate():
    global i
    i = 0
    AsyncObject.cls_value = 0

    yield increment_value.asynq()
    assert_eq(1, i)

    yield increment_value.asynq(), increment_value.asynq(1)
    assert_eq(2, i)

    obj = AsyncObject()
    yield obj.increment_value_method.asynq(), obj.increment_value_method.asynq(1)
    assert_eq(1, obj.value)

    yield AsyncObject.deduplicated_static_method.asynq(), \
        AsyncObject.deduplicated_static_method.asynq(1)
    assert_eq(1, AsyncObject.cls_value)

    i = 0
    yield recursive_call_with_dirty.asynq()

    yield call_with_dirty.asynq()

    if sys.version_info >= (3, 0):
        with AssertRaises(TypeError):
            yield call_with_kwonly_arg.asynq(1)
        assert_eq(1, (yield call_with_kwonly_arg.asynq(arg=1)))


def test_deduplicate_recursion():
    _check_deduplicate_recursion()


@asynq()
def _check_deduplicate_recursion():
    yield recursive_incrementer.asynq(20), increment_value.asynq(0)


def test_async_timer():
    _check_async_timer()


@asynq()
def _slow_task(t):
    yield None
    time.sleep(t)
    result(0); return


@asynq()
def _timed_slow_task(t):
    with AsyncTimer() as timer:
        yield None
        time.sleep(t)
    result(timer.total_time); return


@asynq()
def _check_async_timer():
    with AsyncTimer() as t:
        results = yield [_slow_task.asynq(0.1), _timed_slow_task.asynq(0.1),
                         _slow_task.asynq(0.1), _timed_slow_task.asynq(0.1)]
        assert_eq(0, results[0])
        assert_eq(105000, results[1], tolerance=5000)
        assert_eq(0, results[0])
        assert_eq(105000, results[3], tolerance=5000)

    assert_eq(210000, sum(results), tolerance=10000)
    assert_eq(410000, t.total_time, tolerance=10000)
    assert_gt(t.total_time, sum(results))


def test_async_event_hook():
    calls = []
    @asynq()
    def handler1(*args):
        assert_gt(len(args), 0)
        calls.append('handler1%s' % str(args))

    def handler2(*args):
        calls.append('handler2%s' % str(args))

    hook = AsyncEventHook([handler1])
    hook.subscribe(handler2)

    # trigger
    hook.trigger(1, 2, 'a')
    assert_unordered_list_eq(['handler1(1, 2, \'a\')', 'handler2(1, 2, \'a\')'], calls)

    calls = []
    @asynq()
    def async_trigger():
        yield hook.trigger.asynq(2,3)

    async_trigger()
    assert_unordered_list_eq(['handler1(2, 3)', 'handler2(2, 3)'], calls)

    # safe_trigger
    calls = []
    hook2 = AsyncEventHook([handler1, handler2])
    # calling it with no args will raise AssertionError in handler1
    with AssertRaises(AssertionError):
        hook2.safe_trigger()
    assert_eq(['handler2()'], calls)

    # make sure that the order doesn't matter
    calls = []
    hook3 = AsyncEventHook([handler2, handler1])
    # calling it with no args will raise AssertionError in handler1
    with AssertRaises(AssertionError):
        hook3.safe_trigger()
    assert_eq(['handler2()'], calls)


class DeduplicateClassWrapper:
    @deduplicate()
    @asynq()
    def return_three(self):
        result(3); return

    @deduplicate()
    @asynq()
    def return_five(self):
        result(5); return

    @asynq()
    def return_three_and_five(self):
        result((yield (
            self.return_three.asynq(), self.return_five.asynq()
        ))); return


def test_deduplicate_same_class():
    obj = DeduplicateClassWrapper()

    # make sure the five method has a seperate key and therefore there was no cache mixup
    assert_eq((3, 5), obj.return_three_and_five())



import threading


class LocalProfileState(threading.local):
    def __init__(self):
        super(LocalProfileState, self).__init__()
        self.stats = []
        self.counter = 0


_state = LocalProfileState()
globals()['_state'] = _state


def flush():
    out = _state.stats[:]
    reset()
    return out


def append(stats):
    _state.stats.append(stats)


def reset():
    _state.stats[:] = []
    _state.counter = 0


def incr_counter():
    _state.counter = _state.counter + 1
    return _state.counter



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asynq
from qcore.asserts import assert_eq, assert_is, assert_not_in, AssertRaises

# ===================================================
# Objects to mock
# ===================================================


@asynq.asynq()
def fn():
    pass


@asynq.asynq()
def async_caller():
    ret = yield fn.asynq()
    asynq.result((ret, fn())); return


def non_async_caller():
    return fn()


class Cls(object):
    @asynq.asynq()
    @classmethod
    def async_classmethod(cls):
        pass

    def non_async_method(self, foo):
        pass

    @asynq.asynq()
    def async_method(self):
        return 'capybaras'

instance = Cls()


@asynq.asynq()
def class_method_async_caller():
    ret = yield Cls.async_classmethod.asynq()
    asynq.result((ret, Cls.async_classmethod())); return


def class_method_non_async_caller():
    return Cls.async_classmethod()


@asynq.asynq()
def method_async_caller():
    obj = Cls()
    ret = yield obj.async_method.asynq()
    asynq.result((ret, obj.async_method())); return


def method_non_async_caller():
    return Cls().async_method()


class Counter(object):
    linked_class = Cls

    def __init__(self):
        self.count = 0

    def add_call(self):
        self.count += 1


class SecondClass(object):
    pass


IMPORTANT_DICTIONARY = {'capybaras': 'guinea pigs', 'hutias': 'coypus'}

# ===================================================
# Helpers for testing that mocks work right.
# ===================================================


class MockChecker(object):
    @classmethod
    def check(cls, mock_fn, mock_classmethod, mock_method):
        cls._check_mock(mock_fn, async_caller, non_async_caller)
        cls._check_mock(mock_classmethod, class_method_async_caller, class_method_non_async_caller)
        cls._check_mock(mock_method, method_async_caller, method_non_async_caller)

    @classmethod
    def _check_mock(cls, mock_fn, async_caller, non_async_caller):
        raise NotImplementedError


class MockCheckerWithAssignment(MockChecker):
    @classmethod
    def _check_mock(cls, mock_fn, async_caller, non_async_caller):
        mock_fn.return_value = 42
        assert_eq(0, mock_fn.call_count)
        assert_eq(42, non_async_caller())
        assert_eq(1, mock_fn.call_count)

        assert_eq((42, 42), async_caller())
        assert_eq(3, mock_fn.call_count)

        mock_fn.side_effect = lambda: 43
        assert_eq(43, non_async_caller())
        assert_eq((43, 43), async_caller())


class MockCheckerWithNew(MockChecker):
    @classmethod
    def _check_mock(cls, mock_fn, async_caller, non_async_caller):
        assert_eq(42, non_async_caller())
        assert_eq((42, 42), async_caller())


# ===================================================
# Actual tests.
# ===================================================


def test_mock_async_context():
    with asynq.mock.patch('asynq.tests.test_mock.fn') as mock_fn, \
            asynq.mock.patch.object(Cls, 'async_classmethod') as mock_classmethod, \
            asynq.mock.patch.object(Cls, 'async_method') as mock_method:
        MockCheckerWithAssignment.check(mock_fn, mock_classmethod, mock_method)

    with asynq.mock.patch('asynq.tests.test_mock.fn', lambda: 42) as mock_fn, \
            asynq.mock.patch.object(Cls, 'async_classmethod', classmethod(lambda _: 42)) as mock_classmethod, \
            asynq.mock.patch.object(Cls, 'async_method', lambda _: 42) as mock_method:
        MockCheckerWithNew.check(mock_fn, mock_classmethod, mock_method)


@asynq.mock.patch('asynq.tests.test_mock.fn')
@asynq.mock.patch.object(Cls, 'async_classmethod')
@asynq.mock.patch('asynq.tests.test_mock.Cls.async_method')
def test_mock_async_decorator(mock_method, mock_classmethod, mock_fn):
    MockCheckerWithAssignment.check(mock_fn, mock_classmethod, mock_method)


@asynq.mock.patch('asynq.tests.test_mock.fn')
@asynq.mock.patch.object(Cls, 'async_classmethod')
@asynq.mock.patch('asynq.tests.test_mock.Cls.async_method')
class TestMockAsyncClassDecorator(object):
    def test(self, mock_method, mock_classmethod, mock_fn):
        MockCheckerWithAssignment.check(mock_fn, mock_classmethod, mock_method)


@asynq.mock.patch('asynq.tests.test_mock.fn', lambda: 42)
@asynq.mock.patch.object(Cls, 'async_classmethod', classmethod(lambda _: 42))
@asynq.mock.patch('asynq.tests.test_mock.Cls.async_method', lambda _: 42)
def test_mock_async_decorator_with_new():
    MockCheckerWithNew.check(fn, Cls.async_classmethod, Cls().async_method)


@asynq.mock.patch('asynq.tests.test_mock.fn', lambda: 42)
@asynq.mock.patch.object(Cls, 'async_classmethod', classmethod(lambda _: 42))
@asynq.mock.patch('asynq.tests.test_mock.Cls.async_method', lambda _: 42)
class TestMockAsyncClassDecoratorWithNew(object):
    def test(self):
        MockCheckerWithNew.check(fn, Cls.async_classmethod, Cls().async_method)


def test_mock_async_dict():
    assert_eq('guinea pigs', IMPORTANT_DICTIONARY['capybaras'])

    with asynq.mock.patch.dict(
            'asynq.tests.test_mock.IMPORTANT_DICTIONARY', {'capybaras': 'maras'}):
        assert_eq('maras', IMPORTANT_DICTIONARY['capybaras'])
        assert_eq('coypus', IMPORTANT_DICTIONARY['hutias'])
    assert_eq('guinea pigs', IMPORTANT_DICTIONARY['capybaras'])

    with asynq.mock.patch.dict('asynq.tests.test_mock.IMPORTANT_DICTIONARY', {'capybaras': 'maras'},
                               clear=True):
        assert_eq('maras', IMPORTANT_DICTIONARY['capybaras'])
        assert_not_in('hutias', IMPORTANT_DICTIONARY)


def test_maybe_wrap_new():
    counter = Counter()
    with asynq.mock.patch('asynq.tests.test_mock.fn', counter.add_call):
        fn()
    assert_eq(1, counter.count)

    with asynq.mock.patch('asynq.tests.test_mock.fn', 'capybara'):
        assert_eq('capybara', fn)



class TestPatchMethodWithMethod(object):
    def setup(self):
        self.calls = []

    def mock_method(self, foo):
        self.calls.append(foo)

    def test(self):
        with asynq.mock.patch.object(instance, 'non_async_method', self.mock_method):
            assert_eq([], self.calls)
            instance.non_async_method('bar')
            assert_eq(['bar'], self.calls)


def test_patch_class():
    with asynq.mock.patch.object(Counter, 'linked_class', SecondClass):
        assert_is(SecondClass, Counter.linked_class)


def test_cant_set_attribute():
    with asynq.mock.patch('asynq.tests.test_mock.fn'):
        with AssertRaises(TypeError):
            fn.asynq.cant_set_attribute = 'capybara'



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__doc__ = """

Helpers to use scoped values within async functions.

Similar to AsyncContext in that the scope "pauses" and "resumes"
as the scheduler pauses and resumes the execution of the function
the scoped values are defined/overridden within.

See tests/test_scoped_value.py for examples.

"""

import qcore

from . import contexts


_empty_context = qcore.empty_context


class AsyncScopedValue(object):
    def __init__(self, default):
        self._value = default

    def get(self):
        return self._value

    def set(self, value):
        self._value = value

    def override(self, value):
        """Temporarily overrides the old value with the new one."""
        return _AsyncScopedValueOverrideContext(self, value)

    def __call__(self):
        """Same as get."""
        return self._value

    def __str__(self):
        return 'AsyncScopedValue(%s)' % str(self._value)

    def __repr__(self):
        return 'AsyncScopedValue(%s)' % repr(self._value)


class _AsyncScopedValueOverrideContext(contexts.AsyncContext):
    def __init__(self, target, value):
        self._target = target
        self._value = value
        self._old_value = None

    def resume(self):
        self._old_value = self._target._value
        self._target._value = self._value

    def pause(self):
        self._target._value = self._old_value

    def __repr__(self):
        return '_AsyncScopedValueOverrideContext(target=%r, value=%r)' % (self._target, self._value)


class _AsyncPropertyOverrideContext(contexts.AsyncContext):
    def __init__(self, target, property_name, value):
        self._target = target
        self._property_name = property_name
        self._value = value
        self._old_value = None

    def resume(self):
        self._old_value = getattr(self._target, self._property_name)
        setattr(self._target, self._property_name, self._value)

    def pause(self):
        setattr(self._target, self._property_name, self._old_value)

    def __repr__(self):
        return '_AsyncPropertyOverrideContext(target=%r, property_name=%r, value=%r)' % (
            self._target,
            self._property_name,
            self._value
        )

async_override = _AsyncPropertyOverrideContext
globals()['async_override'] = async_override



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import traceback

import qcore.helpers as core_helpers
import qcore.events as core_events
import qcore.errors as core_errors

from . import debug
from . import _debug


__traceback_hide__ = True


_debug_options = _debug.options
_none = core_helpers.MarkerObject(u"none (futures)")
globals()['_none'] = _none


class FutureIsAlreadyComputed(Exception):
    pass


class FutureBase(object):
    """Abstract base class for delayed computations (AKA
    futures or promises).

    Intended usage:

    * ``future.value()`` gets the value of future computation.
      Computes, if it's not cached yet.
    * ``future()`` does the same (just a shorter notation)

    """
    def __init__(self):
        self._value = _none
        self._error = None
        self._in_repr = False
        self.on_computed = core_events.EventHook()

    def value(self):
        """Gets the value of future.

        Computes, if necessary. If computation was completed
        with exception, it will be raised by this method.

        """
        if self._value is _none:
            self._compute()
        self.raise_if_error()
        return self._value

    def set_value(self, value):
        """Sets cached value and raises on_computed event.
        In addition, sets error property to None.

        """
        if self.is_computed():
            raise FutureIsAlreadyComputed(self)
        self._error = None
        self._value = value
        self._computed()

    def reset_unsafe(self):
        """Resets the cached value making it possible to compute
        the future once more time.

        Normally you should never use this method.

        """
        self._error = None
        self._value = _none

    def error(self):
        """Gets an error that was raised during value computation.

        If computation isn't executed yet, it will be executed by this method.
        Returns None, if there was no error.

        If error is provided, this method calls set_error
        passing provided value to it.

        """
        if self._value is _none:
            self._compute()
        return self._error

    def set_error(self, error):
        """Sets cached error and raises on_computed event.
        In addition, sets value property to None.
        """
        if self.is_computed():
            raise FutureIsAlreadyComputed(self)
        self._error = error
        self._value = None
        self._computed()

    def is_computed(self):
        """Returns true if value and error property values
        were already computed.

        """
        return self._value is not _none

    def _computed(self):
        """Protected method invoked when value or error property
        is set.

        Raises on_computed event.

        """
        if _debug_options.DUMP_COMPUTED:
            debug.write('@async: computed %s' % debug.str(self))
        # Must be a safe call, since we want any subscriber to get
        # the notification even in case one of them fails.
        # Otherwise lots of negative effects are possible - e.g.
        # some of them might be left in blocked state.
        try:
            self.on_computed.safe_trigger(self)

        # pass on any exception in the callback, since we don't expect the caller to know how to
        # deal with it; we don't pass on BaseExceptions (they should be very rare here, and we'd
        # like to maintain their normal behavior), so these could still leave the scheduler in a
        # bad state if the process continues to run afterwards
        except Exception as e:
            print('exception ignored in asynq on_computed callback: %s' % repr(e))
            traceback.print_exc()

    def _compute(self):
        """Protected method invoked to acquire the value.
        Override it to implement value computation.
        It is required to set either value property, or error property
        (in any case both properties will be set).

        """
        raise NotImplementedError()

    def raise_if_error(self):
        if self._error is not None:
            core_errors.reraise(self._error)

    def __call__(self):
        """The same as value property.
        Call arguments must be ignored.

        """
        return self.value()

    def __repr__(self):
        if self._in_repr:
            return '<recursion>'
        try:
            self._in_repr = True
            if self.is_computed():
                status = 'computed, '
                if self.error() is None:
                    if self.value() is self:
                        status += '= self'
                    else:
                        status += '= ' + repr(self.value())
                else:
                    status += 'error = ' + repr(self.error())
            else:
                status = "isn't computed"
            return '%s (%s)' % (type(self), status)
        finally:
            self._in_repr = False

    def dump(self, indent=0):
        debug.write(debug.str(self), indent)

    def __nonzero__(self):
        # treating a FutureBase object as a bool is probably a bug
        raise TypeError("Cannot convert FutureBase object to bool")


class Future(FutureBase):
    """Delegate-based future implementation."""
    def __init__(self, value_provider):
        super(Future, self).__init__()
        self._value_provider = value_provider

    def _compute(self):
        try:
            self.set_value(self._value_provider())
        except Exception as error:
            self.set_error(error)
            raise


class ConstFuture(FutureBase):
    """Future wrapping the constant."""
    def __init__(self, value):
        self._value = _none
        self._error = None
        self.on_computed = core_events.sinking_event_hook  # Simple performance optimization
        self.set_value(value)
        self._in_repr = False  # since we don't call super.__init__

    def __reduce__(self):
        return (ConstFuture, (self._value,))

none_future = ConstFuture(None)


class ErrorFuture(FutureBase):
    """Future that always raises the specified error."""
    def __init__(self, error):
        self._value = _none
        self._error = None
        self.on_computed = core_events.sinking_event_hook  # Simple performance optimization
        self.set_error(error)
        self._in_repr = False  # since we don't call super.__init__



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asynq


class NonAsyncContext(object):
    """Indicates that context can't contain yield statements.

    It means while a NonAsyncContext is active, async tasks cannot
    yield the control back to scheduler.

    Contexts should subclass this class if they want it throw an AssertionError
    if async tasks end up yielding within them.

    Note: Remember to call super inside your __enter__/__exit__.

    """

    def __enter__(self):
        self._active_task = enter_context(self)

    def __exit__(self, typ, val, tb):
        leave_context(self, self._active_task)

    def pause(self):
        assert False, 'Task %s cannot yield while %s is active' % (self._active_task, self)

    def resume(self):
        assert False, 'Task %s cannot yield while %s is active' % (self._active_task, self)


def enter_context(context):
    # perf optimization: inline get_active_task
    active_task = asynq.scheduler._state.current.active_task
    if active_task is not None:
        active_task._enter_context(context)
    return active_task


def leave_context(context, active_task):
    if active_task is not None:
        active_task._leave_context(context)


class AsyncContext(object):
    """Base class for contexts that should pause and resume during an async's function execution.

    Your context should subclass this class and implement pause and resume (at least).

    That would make the context pause and resume each time the execution of the async function
    within this context is paused and resumed.

    Additionally, you can also subclass __enter__ and __exit__ if you want to customize its
    behaviour. Remember to call super in that case.

    NOTE: __enter__/__exit__ methods automatically call resume/pause so the overridden
    __enter__/__exit__ methods shouldn't do that explicitly.

    """

    def __enter__(self):
        self._active_task = enter_context(self)
        self.resume()
        return self

    def __exit__(self, ty, value, tb):
        leave_context(self, self._active_task)
        self.pause()

    def resume(self):
        raise NotImplementedError()

    def pause(self):
        raise NotImplementedError()



import random
import threading
import time

from qcore import SECOND
from qcore.asserts import assert_eq, assert_le, assert_ge
from asynq import asynq, debug, profiler, result
from .debug_cache import reset_caches, mc
from .caching import ExternalCacheBatchItem


class ProfilerThread(threading.Thread):
    def __init__(self, stats_list):
        threading.Thread.__init__(self)
        self.stats_list = stats_list[:]

    def run(self):
        profiler.reset()
        for stats in self.stats_list:
            profiler.append(stats)
        self.profiled_result = profiler.flush()


def test_multithreaded_profiler():
    """Test multithreaded profiler."""
    stats_list = list(range(10000))

    profiler.reset()
    assert_eq([], profiler.flush())

    threads = []
    for i in range(30):
        random.shuffle(stats_list)
        thread = ProfilerThread(stats_list)
        threads.append(thread)

    random.shuffle(stats_list)
    for stats in stats_list:
        profiler.append(stats)

    for thread in threads:
        thread.start()

    for thread in threads:
        thread.join()
        assert_eq(thread.stats_list, thread.profiled_result)

    assert_eq(stats_list, profiler.flush())


def test_collect_perf_stats():
    sleep_time = 50000  # 50ms

    @asynq()
    def func(depth):
        if depth == 0:
            result((yield ExternalCacheBatchItem(mc._batch, 'get', 'test'))); return
        time.sleep(float(sleep_time) / SECOND)
        yield func.asynq(depth - 1), func.asynq(depth - 1)


    debug.options.COLLECT_PERF_STATS = True
    profiler.reset()
    reset_caches()
    depth = 3
    func(depth)
    profiled_result = profiler.flush()
    assert_eq(2 ** (depth + 1) - 1, len(profiled_result))
    for stats in profiled_result:
        num_deps = stats['num_deps']
        num_dependencies = len(stats['dependencies'])
        time_taken = stats['time_taken']

        # It has one ExternalCacheBatchItem, and no dependent AsyncTasks.
        if num_deps == 1:
            assert_eq(0, num_dependencies)
            assert_ge(sleep_time, time_taken)

        # It has two dependent AsyncTasks, and is sleeping for sleep_time (50ms).
        elif num_deps == 2:
            assert_eq(2, num_dependencies)
            assert_le(sleep_time, time_taken)
            assert_ge(sleep_time * 2, time_taken)
        else:
            assert False

    # When COLLET_PERF_STATS is False, we shouldn't profile anything.
    debug.options.COLLECT_PERF_STATS = False
    profiler.reset()
    reset_caches()
    func(2)
    profiled_result = profiler.flush()
    assert_eq(0, len(profiled_result))



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asynq import asynq, result, AsyncContext
from qcore.asserts import AssertRaises, assert_eq

from .helpers import Profiler

counter = 0


def test():
    global counter

    @asynq(pure=True)
    def throw(expected_counter, must_throw):
        global counter
        print("  In throw, counter=%i (expected %i), must_throw=%s" % (counter, expected_counter, str(must_throw)))
        assert expected_counter == counter
        if must_throw:
            raise RuntimeError
        counter += 1
        result(counter); return

    @asynq(pure=True)
    def test():
        global counter
        counter = 0
        tasks = (throw(0, False), throw(1, True), throw(1, False))
        try:
            yield tasks
            raise AssertionError()
        except:
            pass
        assert counter == 2
        assert tasks[0].value() == 1
        assert tasks[1].error() is not None
        assert tasks[2].value() == 2

    with Profiler('test_exceptions()'):
        test()()


context_is_active = 0


def test_async_context():
    class ContextThatRaises(AsyncContext):
        def __init__(self, raise_in_pause, raise_in_resume):
            self.raise_in_pause = raise_in_pause
            self.raise_in_resume = raise_in_resume

        def resume(self):
            if self.raise_in_resume:
                # we raise a BaseException because cythonized async
                # can ignore system generated exceptions as well.
                # e.g. we raise KeyboardInterrupt before restarting a webserver
                raise KeyboardInterrupt()

        def pause(self):
            if self.raise_in_pause:
                raise KeyboardInterrupt()

    class SimpleContext(AsyncContext):
        def resume(self):
            global context_is_active
            context_is_active += 1

        def pause(self):
            global context_is_active
            context_is_active -= 1

    @asynq()
    def dependency():
        return 1

    @asynq()
    def throw(raise_in_pause, raise_in_resume):
        with SimpleContext():
            with ContextThatRaises(raise_in_pause, raise_in_resume):
                with SimpleContext():
                    # we need this to have a real dependency on an async task, otherwise
                    # it executes the whole function inline and the real problem is never tested
                    val = yield dependency.asynq()
        result(val); return

    def check_contexts_released_properly(raise_in_pause, raise_in_resume):
        with AssertRaises(KeyboardInterrupt):
            throw(raise_in_pause, raise_in_resume)

        global context_is_active
        assert_eq(0, context_is_active)

    check_contexts_released_properly(True, False)
    check_contexts_released_properly(False, True)
    check_contexts_released_properly(True, True)



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__doc__ = """

Easy mocking for async functions.

This module provides a patch() function that is based on the standard library's mock.patch but
supports mocking async functions.

This file is named mock_ instead of mock so that it can import the standard mock module.

"""

import inspect

try:
    from unittest import mock
except ImportError:
    # py3
    import mock

try:
    _patch = mock._patch
except AttributeError:
    # some versions of mock put the implementation in mock/mock.py
    _patch = mock.mock._patch
try:
    _get_target = mock._get_target
except AttributeError:
    # some versions of mock put the implementation in mock/mock.py
    _get_target = mock.mock._get_target


from .decorators import asynq
from .futures import ConstFuture


def patch(
        target, new=mock.DEFAULT, spec=None, create=False,
        mocksignature=False, spec_set=None, autospec=False,
        new_callable=None, **kwargs
):
    """Mocks an async function.

    Should be a drop-in replacement for mock.patch that handles async automatically. The .asynq
    attribute is automatically created and shouldn't be used when accessing data on
    the mock.

    """
    getter, attribute = _get_target(target)
    return _make_patch_async(
        getter, attribute, new, spec, create, mocksignature, spec_set, autospec, new_callable,
        kwargs
    )


def _patch_object(
        target, attribute, new=mock.DEFAULT, spec=None,
        create=False, mocksignature=False, spec_set=None, autospec=False,
        new_callable=None, **kwargs
):
    getter = lambda: target
    return _make_patch_async(
        getter, attribute, new, spec, create, mocksignature, spec_set, autospec, new_callable,
        kwargs
    )


patch.object = _patch_object
# duplicate mock.patch.dict for compatibility
patch.dict = mock.patch.dict
patch.TEST_PREFIX = mock.patch.TEST_PREFIX
patch.stopall = mock.patch.stopall


def _make_patch_async(getter, attribute, new, spec, create, mocksignature,
                      spec_set, autospec, new_callable, kwargs):
    new = _maybe_wrap_new(new)

    try:
        patch = _PatchAsync(
            getter, attribute, new, spec, create, mocksignature,
            spec_set, autospec, new_callable, kwargs
        )
    except TypeError:
        # Python 3 doesn't have mocksignature
        patch = _PatchAsync(
            getter, attribute, new, spec, create, spec_set, autospec, new_callable, kwargs
        )
    return patch


class _PatchAsync(_patch):
    def __enter__(self):
        mock_fn = super(_PatchAsync, self).__enter__()
        # so we can also mock non-functions for compatibility
        if callable(mock_fn):
            async_fn = _AsyncWrapper(mock_fn)
            mock_fn.asynq = async_fn
            setattr(mock_fn, 'async', async_fn)
        return mock_fn

    start = __enter__

    def copy(self):
        """Identical to the superclass except that a _PatchAsync object is created."""
        if hasattr(self, 'mocksignature'):  # Python 2
            patcher = _PatchAsync(
                self.getter, self.attribute, self.new, self.spec,
                self.create, self.mocksignature, self.spec_set,
                self.autospec, self.new_callable, self.kwargs
            )
        else:
            patcher = _PatchAsync(
                self.getter, self.attribute, self.new, self.spec,
                self.create, self.spec_set, self.autospec, self.new_callable, self.kwargs
            )
        patcher.attribute_name = self.attribute_name
        patcher.additional_patchers = [
            p.copy() for p in self.additional_patchers
        ]
        return patcher


class _AsyncWrapper(object):
    """Wrapper for the .asynq attribute of patch'ed functions.

    Prevents people from setting and reading attributes on them.

    """

    def __init__(self, mock_fn):
        object.__setattr__(self, '_mock_fn', mock_fn)

    def __call__(self, *args, **kwargs):
        return ConstFuture(self._mock_fn(*args, **kwargs))

    def __setattr__(self, attr, value):
        raise TypeError(
            'You cannot set attributes directly on a .asynq function. Set them on the '
            'function itself instead.'
        )

    def __getattr__(self, attr):
        raise TypeError(
            'You cannot read attributes directly on a .asynq function. Read them on the '
            'function itself instead.'
        )


def _maybe_wrap_new(new):
    """If the mock replacement cannot have attributes set on it, wraps it in a function.

    Also, if the replacement object is a method, applies the async() decorator.

    This is needed so that we support patch(..., x.method) where x.method is an instancemethod
    object, because instancemethods do not support attribute assignment.

    """
    if new is mock.DEFAULT:
        return new

    if inspect.isfunction(new) or isinstance(new, (classmethod, staticmethod)):
        return asynq(sync_fn=new)(new)
    elif not callable(new):
        return new

    try:
        new._maybe_wrap_new_test_attribute = None
        del new._maybe_wrap_new_test_attribute
    except (AttributeError, TypeError):
        # setting something on a bound method raises AttributeError, setting something on a
        # Cythonized class raises TypeError
        should_wrap = True
    else:
        should_wrap = False

    if should_wrap:
        # we can't just use a lambda because that overrides __get__ and creates bound methods we
        # don't want, so we make a wrapper class that overrides __call__
        class Wrapper(object):
            def __call__(self, *args, **kwargs):
                return new(*args, **kwargs)
        return Wrapper()
    else:
        return new



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asynq import AsyncTask, asynq
from asynq.generator import END_OF_GENERATOR, async_generator, list_of_generator, take_first, Value
from qcore.asserts import assert_eq, assert_is, assert_is_instance, AssertRaises


def test_value():
    val = Value('value')
    assert_eq('value', val.value)
    assert_eq("<Value: 'value'>", repr(val))


@asynq()
def alen(seq):
    return len(seq)


@async_generator()
def generator():
    for value in ([], [1], [1, 2]):
        length = yield alen.asynq(value)
        yield Value(length)


@async_generator()
def generator_with_more_yields():
    for task in generator():
        value = yield task
        value = yield alen.asynq([value] * value)
        yield Value(value)
    yield alen.asynq([1, 2])


@async_generator()
def generator_without_yields():
    for i in range(3):
        yield Value(i)


def test_list_of_generator():
    assert_eq([0, 1, 2], list_of_generator(generator()))
    assert_eq([0, 1, 2], list_of_generator(generator_with_more_yields()))


def test_take_first():
    gen = generator()
    assert_eq([0], take_first(gen, 1))
    assert_eq([1, 2], take_first(gen, 2))
    assert_eq([], take_first(gen, 3))

    gen = generator_with_more_yields()
    assert_eq([0, 1, 2], take_first(gen, 4))

    gen = generator_without_yields()
    assert_eq([0, 1, 2], take_first(gen, 4))


def test_must_compute():
    gen = generator()
    next(gen)
    with AssertRaises(RuntimeError):
        next(gen)


def test_values():
    gen = generator_with_more_yields()
    for i in range(3):
        task = next(gen)
        assert_is_instance(task, AsyncTask)
        assert_eq(i, task.value())

    task = next(gen)
    assert_is_instance(task, AsyncTask)
    assert_is(END_OF_GENERATOR, task.value())

    with AssertRaises(StopIteration):
        next(gen)



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asynq import AsyncContext, asynq, debug, result, NonAsyncContext
from .helpers import Profiler
from qcore.asserts import assert_eq, assert_is, AssertRaises

from .debug_cache import mc
from .caching import ExternalCacheBatchItem


current_context = None
change_amount = 0


class Context(AsyncContext):
    def __init__(self, name, parent, assert_state_changes=True):
        self.name = name
        self.parent = parent
        self.state = "pause"
        self.assert_state_changes = assert_state_changes

    def __enter__(self):
        global current_context
        assert_is(self.parent, current_context)
        return super(Context, self).__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        global current_context
        assert_is(self, current_context)
        super(Context, self).__exit__(exc_type, exc_val, exc_tb)

    def resume(self):
        global current_context
        if self.assert_state_changes:
            assert_eq('pause', self.state)
        self.state = "resume"
        current_context = self
        print(self.name + ': resume')

    def pause(self):
        global current_context
        if self.assert_state_changes:
            assert_eq('resume', self.state)
        self.state = "pause"
        current_context = self.parent
        print(self.name + ': pause')


def test_parallel():
    @asynq(pure=True)
    def parallel(name, parent, level):
        with Context(name, parent) as ctx:
            if level >= 2:
                yield debug.sync()
            else:
                yield (
                    parallel(name + '-1', ctx, level + 1),
                    parallel(name + '-2', ctx, level + 1),
                )

    with Profiler('test_parallel()'):
        @asynq()
        def together():
            yield parallel('taskA', None, 0), parallel('taskB', None, 0)
        together()
    print()


def test_adder():
    global change_amount
    change_amount = 0

    @asynq(pure=True)
    def async_add(a, b):
        yield debug.sync()
        z = a + b + change_amount
        result(z); return

    class AsyncAddChanger(AsyncContext):
        def __init__(self, diff):
            self.diff = diff

        def resume(self):
            global change_amount
            change_amount += self.diff

        def pause(self):
            global change_amount
            change_amount -= self.diff

    global expected_change_amount_base
    expected_change_amount_base = 0

    @asynq(pure=True)
    def add_twice(a, b):
        global change_amount
        global expected_change_amount_base
        assert_eq(expected_change_amount_base, change_amount)
        with AsyncAddChanger(1):
            assert_eq(expected_change_amount_base + 1, change_amount)
            z = yield async_add(a, b)
            assert_eq(expected_change_amount_base + 1, change_amount)
            with AsyncAddChanger(1):
                q = yield async_add(a, b)
                assert_eq(expected_change_amount_base + 2, change_amount)
            assert_eq(expected_change_amount_base + 1, change_amount)
        assert_eq(expected_change_amount_base + 0, change_amount)
        result((yield async_add(z, q))); return

    @asynq(pure=True)
    def useless():
        a, b, c = yield (add_twice(1, 1), add_twice(1, 1), async_add(1, 1))
        result((a, b, c)); return

    assert_eq(2, async_add(1, 1)())
    assert_eq((7, 7, 2), useless()())

    expected_change_amount_base += 1
    with AsyncAddChanger(1):
        assert_eq((10, 10, 3), useless()())


class Ctx(NonAsyncContext):
    def __enter__(self):
        super(Ctx, self).__enter__()
        return self

    def __exit__(self, typ, val, tb):
        super(Ctx, self).__exit__(typ, val, tb)
        return


def test_non_async_context():
    @asynq()
    def async_fn_with_yield(should_yield):
        with Ctx():
            if should_yield:
                ret = yield ExternalCacheBatchItem(mc._batch, 'get', 'test')
            else:
                ret = 0
        result(ret); return

    @asynq()
    def batch(should_yield=True):
        ret1, ret2 = yield async_fn_with_yield.asynq(should_yield), \
            async_fn_with_yield.asynq(should_yield)
        result((ret1, ret2)); return

    with AssertRaises(AssertionError):
        batch()

    # assert that without the yields, there is not AssertionError
    batch(False)



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gc

from asynq import asynq, debug, result, AsyncTask
from .helpers import Profiler

values = {}


class WrappedAsyncTask(AsyncTask):
    pass


def wrapped_async(*args, **kwargs):
    return asynq(*args, cls=WrappedAsyncTask, **kwargs)


# async = decorators.async_old
# async = wrapped_async


@asynq(pure=True)
def get(key):
    global values
    result(values.get(key)); return
    yield  # Must be a generator


@asynq(pure=True)
def set(key, value):
    global values
    values[key] = value
    return
    yield  # Must be a generator


@asynq(pure=True)
def get_and_set(key_from, key_to):
    value = yield get(key_from)
    yield set(key_to, value)


@asynq(pure=True)
def performance_test(task_count):
    global values
    values = {}
    assert len(values) == 0  # Nothing is executed yet!

    yield set(0, 0)
    assert len(values) == 1

    yield [get_and_set(i, i + 1) for i in range(0, task_count)]
    assert len(values) == task_count + 1  # Done at this point


def test():
    with Profiler('test_performance(100): warming up'):
        performance_test(100).value()
    gc.collect()
    with Profiler('test_performance(3000): actual test (w/assertions)'):
        performance_test(3000).value()
    gc.collect()
    with debug.disable_complex_assertions(), \
         Profiler('test_performance(3000): actual test (w/o assertions)'):
        performance_test(3000).value()



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cython

cimport contexts


cdef object _empty_context

cdef class AsyncScopedValue(object):
    cpdef public object _value

    cpdef object get(self)
    cpdef set(self, object value)
    cpdef object override(self, object value)

cdef class _AsyncScopedValueOverrideContext(contexts.AsyncContext):
    cdef AsyncScopedValue _target
    cdef object _value
    cdef object _old_value

    cpdef resume(self)
    cpdef pause(self)


cdef class _AsyncPropertyOverrideContext(contexts.AsyncContext):
    cdef object _target
    cdef object _property_name
    cdef object _value
    cdef object _old_value

    cpdef resume(self)
    cpdef pause(self)

cdef object async_override  # Alias of AsyncPropertyOverrideContext



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from qcore import EventHook, miss, none
from asynq import BatchBase, BatchItemBase, ConstFuture


class CacheBase(object):
    '''
    Abstract base class for any cache.
    '''
    def __init__(self):
        self.on_dependency = EventHook()

    def get(self, key):
        '''
        Gets the value from cache.
        Returns miss (as future), if there is no cached value.
        Normally should return a FutureBase descendant -
        local caches are the only exception from this rule
        (they return values directly).
        '''
        raise NotImplementedError()

    def set(self, key, value):
        '''
        Stores the value in the cache.
        If passed value is miss, the value is evicted from cache.
        The actual operation is normally batched,
        so cache implementation should ensure consistency
        for subsequent get operations (i.e. any stored value
        must be instantly available for get operations,
        even although actual operation was batched).
        Normally returns a future returning None or raising
        an error.
        '''
        raise NotImplementedError()

    def _dependency(self, key):
        self.on_dependency(key)


class LocalCache(CacheBase):
    '''
    Local cache implementation (dictionary-based).
    Since all operations can be performed immediately
    by this case, it never returns futures (so it doesn't
    employ batching as well).
    '''
    def __init__(self):
        super(LocalCache, self).__init__()
        self._storage = {}

    def get(self, key):
        self._dependency(key)
        return self._storage.get(key, miss)

    def set(self, key, value):
        self._dependency(key)
        if (value is miss):
            if key in self._storage:
                del self._storage[key]
        else:
            self._storage[key] = value
        return

    def clear(self):
        self._storage.clear()
        return


class ExternalCacheBatch(BatchBase):
    '''
    Batch implementation for external caches.
    '''
    def __init__(self, cache, index):
        super(ExternalCacheBatch, self).__init__()
        self.cache = cache
        self.index = index  # Batch index. Just for debugging purposes.

    def _try_switch_active_batch(self):
        if self.cache._batch is self:
            self.cache._batch = ExternalCacheBatch(self.cache, self.index + (1 if self.items else 0))

    def _flush(self):
        self.cache._before_flush(self)
        self.cache._flush(self)

    def _cancel(self):
        self.cache._cancel_flush(self)


class ExternalCacheBatchItem(BatchItemBase):
    '''
    Batch item implementation for external caches.
    '''
    def __init__(self, batch, operation, *args):
        super(ExternalCacheBatchItem, self).__init__(batch)
        self.operation = operation
        self.args = args

    def __str__(self):
        return '%s (%s)' % (self.operation, ', '.join(map(lambda i: repr(i), self.args)))


class ExternalCacheBase(CacheBase):
    '''
    Abstract base class for external caches (memcached, etc.).
    '''
    def __init__(self):
        super(ExternalCacheBase, self).__init__()
        self._local = LocalCache()
        self._batch = ExternalCacheBatch(self, 1)
        self.use_local_cache = True

    def get(self, key):
        self._dependency(key)
        value = self._local.get(key)
        if value is miss:  # No item is cached locally
            return self._create_batch_item('get', key)
        return ConstFuture(miss if value is none else value)

    def set(self, key, value):
        self._dependency(key)
        if self.use_local_cache:
            self._local.set(key, none if value is miss else value)
        return self._create_batch_item('set', key, value)

    def clear(self):
        self.clear_local_cache()
        return self._create_batch_item('clear')

    def flush(self):
        return self._batch.flush()

    def cancel_flush(self):
        return self._batch.cancel()

    def clear_local_cache(self):
        self._local.clear()

    def _before_flush(self, batch):
        pass

    def _flush(self, batch):
        # Actual batch flush implementation
        raise NotImplementedError()

    def _cancel_flush(self, batch):
        # Actual batch cancel implementation
        pass

    def _create_batch_item(self, operation, *args):
        return ExternalCacheBatchItem(self._batch, operation, *args)



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, division, print_function

import sys
import traceback
import logging

import six
from six import StringIO
from qcore.asserts import assert_eq, assert_in, assert_is, assert_is_not
from qcore import MarkerObject, prepare_for_reraise

import asynq
from asynq import result


def test_dump_error():
    buf = StringIO()

    with asynq.mock.patch('asynq.debug.stderr', buf):
        asynq.debug.dump_error(None)
    assert_eq('\nNo error', buf.getvalue())


@asynq.asynq()
def async_fn():
    pass


def test_format_error():
    assert_is(None, asynq.debug.format_error(None))

    e = RuntimeError()
    expected = '\nRuntimeError\n'
    assert_eq(expected, asynq.debug.format_error(e))

    e._task = async_fn.asynq()
    formatted = asynq.debug.format_error(e)
    assert_in(expected, formatted)

    try:
        raise RuntimeError
    except RuntimeError:
        e._traceback = sys.exc_info()[2]

    formatted = asynq.debug.format_error(e)
    assert_in(expected, formatted)
    assert_in('Traceback', formatted)


def test_format_error_chaining():
    if six.PY2:
        return  # py2 doesn't have chaining

    try:
        try:
            raise ValueError
        except ValueError:
            raise KeyError
    except KeyError as e:
        prepare_for_reraise(e)
        exc = e

    formatted = asynq.debug.format_error(exc)
    assert_in('raise ValueError', formatted)
    assert_in('raise KeyError', formatted)
    assert_in('During handling of the', formatted)


def test_dump_stack():
    buf = StringIO()

    with asynq.mock.patch('sys.stdout', buf):
        def inner():
            asynq.debug.dump_stack()
        inner()

    printed = buf.getvalue()
    assert_in('test_dump_stack', printed)
    assert_in('Stack trace:', printed)


def test_format_asynq_stack():
    format_list = []

    @asynq.asynq()
    def level1(arg):
        if arg == 0:
            format_list.append(asynq.debug.format_asynq_stack())
        result(arg); return

    @asynq.asynq()
    def level2(arg):
        result((yield level1.asynq(arg))); return

    @asynq.asynq()
    def root():
        vals = yield [level2.asynq(i) for i in range(5)]
        result(vals); return

    root()

    traceback = '\n'.join(format_list[0])
    assert_in('root', traceback)
    assert_in('level1', traceback)
    assert_in('level2', traceback)


def _assert_write_result(text, indent, expected):
    buf = StringIO()

    with asynq.mock.patch('asynq.debug.stdout', buf):
        asynq.debug.write(text, indent=indent)

    assert_eq(expected, buf.getvalue())


def test_write():
    _assert_write_result('capybara', 0, 'capybara\n')
    _assert_write_result('capybara', 1, '  capybara\n')
    _assert_write_result('capybara', 2, '    capybara\n')
    _assert_write_result('nutria\ncapybara', 1, '  nutria\n  capybara\n')
    _assert_write_result('', 1, '  \n')
    _assert_write_result('\n', 1, '\n  \n')
    _assert_write_result('\ncapybara', 1, '\n  capybara\n')


def non_async_function_that_raises_an_error():
    raise ValueError


@asynq.asynq()
def async_function_that_raises_an_error():
    yield
    non_async_function_that_raises_an_error()


@asynq.asynq()
def async_function_whose_child_async_task_will_throw_an_error():
    yield async_function_that_raises_an_error.asynq()


def test_asynq_traceback_gets_glued_at_each_task_level():
    # tests that exceptions are producting the right tracebacks
    traceback_to_verify = None
    try:
        async_function_whose_child_async_task_will_throw_an_error()
    except ValueError:
        traceback_to_verify = sys.exc_info()[2]
    assert_is_not(None, traceback_to_verify)
    traceback_printed = '\n'.join(traceback.format_tb(traceback_to_verify))
    assert_in(non_async_function_that_raises_an_error.__name__, traceback_printed)
    assert_in(async_function_that_raises_an_error.__name__, traceback_printed)
    assert_in(async_function_whose_child_async_task_will_throw_an_error.__name__, traceback_printed)


def assert_eq_extracted_traceback_entry(
        entry,
        filename,
        fn_name,
        line):
    entry_filename, _, entry_fn_name, entry_line = entry
    assert_in(filename.rstrip('c'), entry_filename)
    assert_eq(fn_name, entry_fn_name)
    assert_eq(line, entry_line)


def test_extract_traceback():
    traceback_to_verify = None
    try:
        async_function_whose_child_async_task_will_throw_an_error()
    except ValueError:
        traceback_to_verify = sys.exc_info()[2]
    extracted_traceback_to_verify = asynq.debug.extract_tb(
            traceback_to_verify)

    this_level = extracted_traceback_to_verify[0]
    assert_eq_extracted_traceback_entry(
        this_level,
        __file__,
        'test_extract_traceback',
        'async_function_whose_child_async_task_will_throw_an_error()')

    # now check the last 3 frames. Between the 1st and the last 3 frames there may
    # or may not be a frame from six.reraise depending on the python version.
    async_wrapper_level = extracted_traceback_to_verify[-3]
    assert_eq_extracted_traceback_entry(
        async_wrapper_level,
        __file__,
        'async_function_whose_child_async_task_will_throw_an_error',
        'yield async_function_that_raises_an_error.asynq()')

    async_raiser_level = extracted_traceback_to_verify[-2]
    assert_eq_extracted_traceback_entry(
        async_raiser_level,
        __file__,
        'async_function_that_raises_an_error',
        'non_async_function_that_raises_an_error()')

    normal_function_raising_exception_level = extracted_traceback_to_verify[-1]
    assert_eq_extracted_traceback_entry(
        normal_function_raising_exception_level,
        __file__,
        'non_async_function_that_raises_an_error',
        'raise ValueError')


a_return_value = MarkerObject(u'A return value.')


@asynq.mock.patch('asynq.debug.traceback.format_list')
@asynq.mock.patch('asynq.debug.extract_tb')
def test_format_tb(mock_extract_tb, mock_format_list):
    mock_extract_tb.return_value = a_return_value
    mock_format_list.side_effect = lambda arg: arg
    traceback_to_verify = None
    try:
        async_function_whose_child_async_task_will_throw_an_error()
    except ValueError:
        traceback_to_verify = sys.exc_info()[2]
    assert_is(a_return_value, asynq.debug.format_tb(traceback_to_verify))
    mock_extract_tb.assert_called_once_with(traceback_to_verify)
    mock_format_list.assert_called_once_with(a_return_value)


@asynq.mock.patch('asynq.debug.format_error')
def test_asynq_stack_trace_formatter(mock_format_error):
    mock_format_error.return_value = u'This is some traceback.'
    stderr_string_io = StringIO()
    handler = logging.StreamHandler(stream=stderr_string_io)
    handler.setFormatter(asynq.debug.AsynqStackTracebackFormatter())
    logger = logging.getLogger('test_asynq')
    logger.addHandler(handler)
    exc_info = None
    try:
        async_function_whose_child_async_task_will_throw_an_error()
    except ValueError:
        exc_info = sys.exc_info()
        logger.exception('Test')
    ty, val, tb = exc_info
    mock_format_error.assert_called_once_with(val, tb=tb)
    assert_eq(u'Test\nThis is some traceback.\n', stderr_string_io.getvalue())



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from qcore import none, miss
from asynq import scheduler
from .caching import ExternalCacheBase, LocalCache


# Caches

class DebugExternalCache(ExternalCacheBase):
    def __init__(self, name):
        super(DebugExternalCache, self).__init__()
        self.name = name
        self._remote = LocalCache()
        self.run_scheduler_before_flush = False

    def _flush(self, batch):
        print('%s batch %i:' % (self.name, batch.index))
        for item in batch.items:
            print('  %s' % str(item))
            if item.is_computed():
                # Item can be computed due to task cancellation
                continue
            try:
                method = self.__getattribute__('_flush_' + item.operation)
                item.set_value(method(item, *item.args))
            except BaseException as e:
                item.set_error(e)
                # No need to re-throw, since user will anyway
                # see it by accessing the value

    def _cancel_flush(self, batch):
        print('%s batch %i is cancelled.' % (self.name, batch.index))

    def _flush_get(self, item, key):
        value = self._remote.get(key)
        if self.use_local_cache:
            self._local.set(key, none if value is miss else value)
        return value

    def _flush_set(self, item, key, value):
        self._remote.set(key, value)
        return

    def _flush_clear(self, item):
        self._remote.clear()
        return


mc = DebugExternalCache('MC')


def reset_caches():
    print('Cache reset:')
    mc.cancel_flush()
    mc.clear()
    flush_caches()
    reset_cache_batch_indexes()
    print('Caches are clear.')
    print('')


def reset_cache_batch_indexes():
    mc._batch.index = 1


def flush_caches():
    if mc._batch.items:
        mc.flush()


def flush_and_clear_local_caches():
    flush_caches()
    mc._local.clear()
    db._local.clear()
